select {
  -moz-appearance: none;
}

.ref {
  background-color: green;
}

.enabled {
  background-color: red;
}

:enabled.enabled {
  background-color: green;
}

.disabled {
  background-color: green;
}

:enabled.disabled {
  background-color: red;
}
