/* This Source Code Form is subject to the terms of the Mozilla Public
 * License, v. 2.0. If a copy of the MPL was not distributed with this
 * file, You can obtain one at http://mozilla.org/MPL/2.0/. */

/* This sheet is added to the style set for documents with frames disabled */

noframes {
  display: block;
}

frame, frameset, iframe {
  display: none !important;
}
